import json
from typing import Optional
from pydantic import BaseModel

from holmes.core.tools import StructuredToolResult, StructuredToolResultStatus


class TruncationMetadata(BaseModel):
    tool_call_id: str
    start_index: int
    end_index: int
    tool_name: str
    original_token_count: int


class TruncationResult(BaseModel):
    truncated_messages: list[dict]
    truncations: list[TruncationMetadata]


def format_tool_result_data(tool_result: StructuredToolResult) -> str:
    tool_response = tool_result.data
    if isinstance(tool_result.data, str):
        tool_response = tool_result.data
    else:
        try:
            if isinstance(tool_result.data, BaseModel):
                tool_response = tool_result.data.model_dump_json(indent=2)
            else:
                tool_response = json.dumps(tool_result.data, indent=2)
        except Exception:
            tool_response = str(tool_result.data)
    if tool_result.status == StructuredToolResultStatus.ERROR:
        tool_response = f"{tool_result.error or 'Tool execution failed'}:\n\n{tool_result.data or ''}".strip()
    return tool_response


class ToolCallResult(BaseModel):
    tool_call_id: str
    tool_name: str
    description: str
    result: StructuredToolResult
    size: Optional[int] = None

    def as_tool_call_message(self):
        content = format_tool_result_data(self.result)
        if self.result.params:
            content = (
                f"Params used for the tool call: {json.dumps(self.result.params)}. The tool call output follows on the next line.\n"
                + content
            )
        return {
            "tool_call_id": self.tool_call_id,
            "role": "tool",
            "name": self.tool_name,
            "content": content,
        }

    def as_tool_result_response(self):
        result_dump = self.result.model_dump()
        result_dump["data"] = self.result.get_stringified_data()

        return {
            "tool_call_id": self.tool_call_id,
            "tool_name": self.tool_name,
            "description": self.description,
            "role": "tool",
            "result": result_dump,
        }

    def as_streaming_tool_result_response(self):
        result_dump = self.result.model_dump()
        result_dump["data"] = self.result.get_stringified_data()

        return {
            "tool_call_id": self.tool_call_id,
            "role": "tool",
            "description": self.description,
            "name": self.tool_name,
            "result": result_dump,
        }



from holmes.common.env_vars import TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT
from holmes.core.llm import LLM
from holmes.core.tools import StructuredToolResultStatus
from holmes.core.tools_utils.data_types import ToolCallResult
from holmes.utils import sentry_helper


def prevent_overly_big_tool_response(tool_call_result: ToolCallResult, llm: LLM):
    if (
        tool_call_result.result.status == StructuredToolResultStatus.SUCCESS
        and 0 < TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT
        and TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT <= 100
    ):
        message = tool_call_result.as_tool_call_message()

        messages_token = llm.count_tokens_for_message(messages=[message])
        context_window_size = llm.get_context_window_size()
        max_tokens_allowed: int = int(
            context_window_size * TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT // 100
        )

        if messages_token > max_tokens_allowed:
            relative_pct = (
                (messages_token - max_tokens_allowed) / messages_token
            ) * 100
            error_message = f"The tool call result is too large to return: {messages_token} tokens.\nThe maximum allowed tokens is {max_tokens_allowed} which is {format(relative_pct, '.1f')}% smaller.\nInstructions for the LLM: try to repeat the query but proactively narrow down the result so that the tool answer fits within the allowed number of tokens."
            tool_call_result.result.status = StructuredToolResultStatus.ERROR
            tool_call_result.result.data = None
            tool_call_result.result.error = error_message

            sentry_helper.capture_toolcall_contains_too_many_tokens(
                tool_call_result, messages_token, max_tokens_allowed
            )
